import numpy as np
from acoustic_toolbox import Signal
import pytest
import tempfile

from acoustic_toolbox.signal import EqualBand
import matplotlib.pyplot as plt


def test_wav():
    """Test writing to and reading from wav file."""
    duration = 5.0
    fs = 10025
    samples = int(fs * duration)
    channels = 3

    signal = Signal(np.random.randn(channels, samples), fs)

    with tempfile.TemporaryFile() as file:
        signal.to_wav(file)
        signal = Signal.from_wav(file)
        assert signal.samples == samples
        assert signal.fs == fs
        assert signal.channels == channels


class TestSignal:
    # (channels, samples, sample rate)
    @pytest.fixture(params=[(1, 88200, 22050), (3, 88200, 22050), (3, 88200, 44100)])
    def signal(self, request):
        return Signal(
            np.random.randn(request.param[0], request.param[1]), request.param[2]
        )

    def test_samples(self, signal):
        x = signal.samples

    def test_calibrate_to_scalar(self, signal):
        # Scalar decibel
        signal.calibrate_to(100.0)
        signal.copy().calibrate_to(100.0, inplace=True)

    def test_calibrate_to_channels(self, signal):
        # Value per channel. Note that [...,None] is required!
        signal.calibrate_to((np.ones(signal.channels) * 100.0)[..., None])
        signal.copy().calibrate_to(
            (np.ones(signal.channels) * 100.0)[..., None], inplace=True
        )

    def test_calibrate_to_samples(self, signal):
        # Value per samples
        signal.calibrate_to(np.ones(signal.samples))
        signal.copy().calibrate_to(np.ones(signal.samples), inplace=True)

    def test_calibrate_to_samples_channels(self, signal):
        # Value per sample per channel
        signal.calibrate_to(np.ones(signal.shape))
        signal.copy().calibrate_to(np.ones(signal.shape), inplace=True)

    def test_calibrate_with(self, signal):
        calibration_signal_level = 50.0
        decibel = 94.0
        calibration_signal = Signal(
            np.random.randn(signal.samples), signal.fs
        ).calibrate_to(calibration_signal_level)

        out = signal.calibrate_with(calibration_signal, decibel)
        assert (
            (out.leq() - signal.leq()).mean() - (decibel - calibration_signal_level)
        ) < 0.01

    def test_channels(self, signal):
        x = signal.channels

    def test_duration(self, signal):
        x = signal.duration

    def test_decimate(self, signal):
        factor = 4
        decimated = signal.decimate(factor)
        assert signal.fs / factor == decimated.fs

    def test_upsample(self, signal):
        factor = 2
        assert (signal.upsample(factor).fs / signal.fs) == factor

    def test_gain_scalar(self, signal):
        gain = +20.0
        # `.all()` because of multichannel signals
        assert (np.abs(signal.gain(gain).leq() - (signal.leq() + gain)) < 0.01).all()
        assert (
            np.abs(signal.copy().gain(gain, inplace=True).leq() - (signal.leq() + gain))
            < 0.01
        ).all()

    def test_pick(self, signal):
        x = signal.pick(signal.duration * 0.1, signal.duration * 0.6)

    def test_times(self, signal):
        times = signal.times()

    def test_energy(self, signal):
        energy = signal.energy()

    def test_power(self, signal):
        power = signal.power()

    def test_ms(self, signal):
        ms = signal.ms()

    def test_rms(self, signal):
        rms = signal.rms()

    def test_correlate(self, signal):
        signal = signal[..., 0:100]
        if signal.channels > 1:  # Multichannel is not supported
            with pytest.raises(ValueError):
                assert (signal.correlate() == signal.correlate(signal)).all()
        else:
            assert (signal.correlate() == signal.correlate(signal)).all()

    def test_amplitude_envelope(self, signal):
        x = signal.amplitude_envelope()

    def test_instantaneous_frequency(self, signal):
        x = signal.instantaneous_frequency()

    def test_instantaneous_phase(self, signal):
        x = signal.instantaneous_phase()

    def test_detrend(self, signal):
        x = signal.detrend()

    def test_unwrap(self, signal):
        x = signal.unwrap()

    def test_complex_cepstrum(self, signal):
        t, c, d = signal.complex_cepstrum()

    def test_real_cepstrum(self, signal):
        t, c = signal.real_cepstrum()

    def test_power_spectrum(self, signal):
        freq, power = signal.power_spectrum()

    def test_phase_spectrum(self, signal):
        freq, phase = signal.phase_spectrum()

    def test_peak(self, signal):
        value = signal.peak()
        assert len(value) == signal.channels

    def test_peak_level(self, signal):
        value = signal.peak_level()
        assert len(value) == signal.channels

    def test_sound_exposure(self, signal):
        value = signal.sound_exposure()
        assert len(value) == signal.channels

    def test_sound_exposure_level(self, signal):
        value = signal.sound_exposure_level()
        assert len(value) == signal.channels

    def test_octaves(self, signal):
        freq, octaves = signal.octaves()

    def test_levels(self, signal):
        times, levels = signal.levels()

    def test_leq(self, signal):
        # s = Signal(np.random.randn(10000), 22050)

        leq = signal.leq()

        assert type(leq) is np.ndarray

    def test_bandpass(self, signal):
        x = signal.bandpass(1000.0, 2000.0)

    def test_bandstop(self, signal):
        x = signal.bandstop(1000.0, 2000.0)

    def test_highpass(self, signal):
        x = signal.highpass(1000.0)

    def test_lowpass(self, signal):
        x = signal.lowpass(1000.0)

    def test_octavepass(self, signal):
        x = signal.octavepass(1000.0, fraction=6)

    def test_bandpass_frequencies(self, signal):
        f = EqualBand(center=[100.0, 200.0, 300.0], bandwidth=20.0)
        f, x = signal.bandpass_frequencies(f)

    def test_bandpass_octaves(self, signal):
        f, x = signal.octaves()

    def test_bandpass_third_octaves(self, signal):
        f, x = signal.third_octaves()

    def test_bandpass_fractional_octaves(self, signal):
        f, x = signal.fractional_octaves()

    def test_weigh(self, signal):
        s = signal.weigh()
        s = signal.weigh("C")
        s = signal.weigh("A", zero_phase=True)

    ## Plot methods with arguments to test.
    # plot_methods = {'plot'                      : None,
    #'plot_levels'               :   {
    #'time'                  : [None, 0.125, 1.0],
    #'method'                : ['average', 'weighting'],
    # },
    #'plot_octaves'              : None,
    #'plot_third_octaves'        : None,
    #'plot_fractional_octaves'   : {
    #'fraction'              : [3, 6]
    # },
    #'plot_spectrum'             : {
    #'N'                     : [None, 8000]
    # },
    # }

    # @pytest.yield_fixture
    # def plot_function_with_argument(self):
    ## This won't work with pytest. Apparently they do teardown after the yield
    ## statement and therefore don't support multiple yield statements.
    ## Using a closure doesn't help either.
    # for func, arguments in self.plot_methods.items():
    # if arguments is not None:
    # for prod in it.product(*arguments.values()):
    # yield (func, dict(zip(arguments.keys(), prod)))
    # else:
    # yield (func, None)

    # def test_plot_functions(self, signal, plot_function_with_argument):
    # func, arguments = plot_function_with_argument
    # if arguments is None:
    # getattr(signal, func)()
    # else:
    # getattr(signal, func)(**arguments)

    def test_plot(self, signal):
        signal.plot()
        plt.close("all")

    def test_plot_levels(self, signal):
        signal.plot_levels()
        signal.plot_levels(method="average", time=1.0)
        signal.plot_levels(method="weighting", time=1.0)
        plt.close("all")

    def test_plot_octaves(self, signal):
        signal.plot_octaves()
        plt.close("all")

    def test_plot_third_octaves(self, signal):
        signal.plot_third_octaves()
        plt.close("all")

    def test_plot_fractional_octaves(self, signal):
        signal.plot_fractional_octaves(3)
        signal.plot_fractional_octaves(6)
        signal.plot_fractional_octaves(9)
        plt.close("all")

    def test_plot_power_spectrum(self, signal):
        signal.plot_power_spectrum()
        plt.close("all")

    def test_plot_phase_spectrum(self, signal):
        signal.plot_phase_spectrum()
        plt.close("all")

    def test_plot_spectrogram(self, signal):
        if signal.channels > 1:
            with pytest.raises(ValueError):
                signal.plot_spectrogram()
        else:
            try:
                signal.plot_spectrogram()
            except (
                NotImplementedError
            ):  # easy way to skip mpl 1.3.1 specgram mode issue
                pass
        plt.close("all")

    def spectrogram(self, signal):
        signal.spectrogram()

    def test_pickling(self, signal):
        import pickle

        p = pickle.dumps(signal)
        obj = pickle.loads(p)

        assert (obj == signal).all()
        assert obj.fs == signal.fs
        assert type(obj) is type(signal)



"""The bands module contains functions for working with octave and third-octave frequency bands, based on [IEC 61672-1:2013](standards/iec_61672_1_2013.md).

Functions:
    octave: Generate center frequencies for octave bands.
    octave_low: Calculate lower corner frequencies for octave bands.
    octave_high: Calculate upper corner frequencies for octave bands.
    third: Generate center frequencies for third-octave bands.
    third_low: Calculate lower corner frequencies for third-octave bands.
    third_high: Calculate upper corner frequencies for third-octave bands.
    third2oct: Convert third-octave band levels to octave band levels.
"""

import numpy as np
from typing import Literal
from numpy.typing import NDArray

# from acoustic_toolbox.decibel import dbsum
import acoustic_toolbox
from acoustic_toolbox.standards.iec_61672_1_2013 import (
    NOMINAL_OCTAVE_CENTER_FREQUENCIES,
    NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
)

OCTAVE_CENTER_FREQUENCIES = NOMINAL_OCTAVE_CENTER_FREQUENCIES
"""Preferred nominal octave band center frequencies."""

THIRD_OCTAVE_CENTER_FREQUENCIES = NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES
"""Preferred nominal third-octave band center frequencies."""


def octave(first: float, last: float) -> NDArray[np.float64]:
    """Generate a Numpy array for central frequencies of octave bands.

    There are more information on how to calculate 'real' bands in
    http://blog.prosig.com/2006/02/17/standard-octave-bands/

    Args:
        first: First octave centerfrequency.
        last: Last octave centerfrequency.

    Returns:
        Array of octave band center frequencies in Hz.
    """
    # octave_bands = OCTAVE_CENTER_FREQUENCIES
    # low = np.where(octave_bands == first)[0]
    # high = np.where(octave_bands == last)[0]
    # return octave_bands[low: high+1]
    return acoustic_toolbox.signal.OctaveBand(
        fstart=first, fstop=last, fraction=1
    ).nominal


def octave_low(first: float, last: float) -> NDArray[np.float64]:
    """Lower cornerfrequencies of octaves.

    Args:
        first: First octave centerfrequency.
        last: Last octave centerfrequency.

    Returns:
        Array of lower corner frequencies in Hz.
    """
    return octave(first, last) / np.sqrt(2.0)
    # return acoustic_toolbox.signal.OctaveBand(fstart=first, fstop=last, fraction=1).lower


def octave_high(first: float, last: float) -> NDArray[np.float64]:
    """Upper cornerfrequencies of octaves.

    Args:
        first: First octave centerfrequency.
        last: Last octave centerfrequency.

    Returns:
        Array of upper corner frequencies in Hz.
    """
    return octave(first, last) * np.sqrt(2.0)
    # return acoustic_toolbox.signal.OctaveBand(fstart=first, fstop=last, fraction=1).upper


def third(first: float, last: float) -> NDArray[np.float64]:
    """Generate a Numpy array for central frequencies of third octave bands.

    Args:
        first: First third octave centerfrequency.
        last: Last third octave centerfrequency.

    Returns:
        Array of third octave band center frequencies in Hz.
    """
    # third_oct_bands = THIRD_OCTAVE_CENTER_FREQUENCIES
    # low = np.where(third_oct_bands == first)[0]
    # high = np.where(third_oct_bands == last)[0]
    # return third_oct_bands[low: high+1]
    return acoustic_toolbox.signal.OctaveBand(
        fstart=first, fstop=last, fraction=3
    ).nominal


def third_low(first: float, last: float) -> NDArray[np.float64]:
    """Lower cornerfrequencies of third-octaves.

    Args:
        first: First third octave centerfrequency.
        last: Last third octave centerfrequency.

    Returns:
        Array of lower corner frequencies in Hz.
    """
    return third(first, last) / 2.0 ** (1.0 / 6.0)
    # return acoustic_toolbox.signal.OctaveBand(fstart=first, fstop=last, fraction=3).lower


def third_high(first: float, last: float) -> NDArray[np.float64]:
    """Higher cornerfrequencies of third-octaves.

    Args:
        first: First third octave centerfrequency.
        last: Last third octave centerfrequency.

    Returns:
        Array of upper corner frequencies in Hz.
    """
    return third(first, last) * 2.0 ** (1.0 / 6.0)
    # return Octaveband(fstart=first, fstop=last, fraction=3).upper


def third2oct(
    levels: NDArray[np.float64], axis: int | None = None
) -> NDArray[np.float64]:
    """Calculate Octave levels from third octave levels.

    Args:
        levels: Array containing third octave levels.
        axis: Axis over which to perform the summation.

    Returns:
        Array containing octave band levels.

    Note:
        The number of elements along the summation axis should be a factor of 3. : Third octave levels

    Raises:
        ValueError: If the shape of levels array is not compatible with the operation.
    """
    levels = np.array(levels)
    axis = axis if axis is not None else levels.ndim - 1

    try:
        assert levels.shape[axis] % 3 == 0
    except AssertionError:
        raise ValueError("Wrong shape.")
    shape = list(levels.shape)
    shape[axis] = shape[axis] // 3
    shape.insert(axis + 1, 3)
    levels = np.reshape(levels, shape)
    return np.squeeze(acoustic_toolbox.decibel.dbsum(levels, axis=axis + 1))


def _check_band_type(
    freqs: NDArray[np.float64],
) -> Literal["octave", "octave-unsorted", "third", "third-unsorted"] | None:
    """Check if an array contains octave or third octave bands values sorted or unsorted.

    Args:
        freqs: Array of frequencies to check.

    Returns:
        The type of band ("octave", "octave-unsorted", "third", "third-unsorted") or None if not recognized.
    """
    octave_bands = octave(16, 16000)
    third_oct_bands = third(12.5, 20000)

    def _check_sort(freqs: NDArray[np.float64], bands: NDArray[np.float64]) -> bool:
        """Check if frequencies are in sorted order within the bands.

        Args:
            freqs: Array of frequencies to check.
            bands: Reference band frequencies.

        Returns:
            bool: True if frequencies are in sorted order.
        """
        index = np.where(np.isin(bands, freqs))[0]
        band_pos = index - index[0]
        return (band_pos == np.arange(band_pos.size)).all()

    if np.isin(freqs, octave_bands).all():
        is_sorted = _check_sort(freqs, octave_bands)
        return "octave" if is_sorted else "octave-unsorted"
    elif np.isin(freqs, third_oct_bands).all():
        is_sorted = _check_sort(freqs, third_oct_bands)
        return "third" if is_sorted else "third-unsorted"
    else:
        return None



import pytest
import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal
from acoustic_toolbox.descriptors import leq, sel, lw, lden, ldn


@pytest.fixture
def levels():
    return [80, 80.3, 80.6, 80.4, 79.7, 76.9]


@pytest.fixture
def lday():
    return np.array([80, 70])


@pytest.fixture
def levening():
    return np.array([65, 55])


@pytest.fixture
def lnight():
    return np.array([45, 55])


@pytest.fixture
def power():
    return np.array([1e-3, 1e-5, 1e-7])


class TestDescriptors:
    """Test :mod:`acoustic_toolbox.descriptors`"""

    def test_leq(self, levels):
        calculated = leq(levels)
        real = 79.806989166
        assert_almost_equal(calculated, real)

    def test_leq_int_time(self, levels):
        calculated = leq(levels, int_time=1 / 8)
        real = 88.837889036
        assert_almost_equal(calculated, real)

    def test_sel(self, levels):
        calculated = sel(levels)
        real = 87.588501670
        assert_almost_equal(calculated, real)

    def test_lw_int(self, power):
        calculated = lw(power[0])
        real = 90
        assert calculated == real

    def test_lw_1darray(self, power):
        calculated = lw(power)
        real = np.array([90, 70, 50])
        assert_array_almost_equal(calculated, real)

    def test_lw_1darray_wref(self, power):
        calculated = lw(power, Wref=1e-11)
        real = np.array([80, 60, 40])
        assert_array_almost_equal(calculated, real)

    def test_lden_float(self, lday, levening, lnight):
        calculated = lden(lday[0], levening[0], lnight[0])
        real = 77.14095579
        assert_almost_equal(calculated, real)

    def test_lden_array(self, lday, levening, lnight):
        calculated = lden(lday, levening, lnight)
        real = np.array([77.14095579, 67.93843392])
        assert_array_almost_equal(calculated, real)

    def test_ldn_float(self, lday, lnight):
        calculated = ldn(lday[0], lnight[0])
        real = 77.96703252
        assert_almost_equal(calculated, real)

    def test_ldn_array(self, lday, lnight):
        calculated = ldn(lday, lnight)
        real = np.array([77.96703252, 68.71330861])
        assert_array_almost_equal(calculated, real)



import numpy as np
from acoustic_toolbox.atmosphere import Atmosphere

from tests.get_data_path import data_path


class TestAtmosphere:
    """Test :mod:`acoustic_toolbox.atmosphere`"""

    def test_standard_atmosphere(self):
        a = Atmosphere()
        """Default values."""
        assert a.temperature == 293.15
        assert a.pressure == 101.325
        assert a.relative_humidity == 0.0
        """Calculated values belonging to default values."""
        assert abs(a.soundspeed - 343.2) < 1.0e-9
        assert abs(a.saturation_pressure - 2.33663045) < 1.0e-8
        assert abs(a.molar_concentration_water_vapour - 0.0) < 1.0e-9
        assert abs(a.relaxation_frequency_nitrogen - 9.0) < 1.0e-9
        assert abs(a.relaxation_frequency_oxygen - 24.0) < 1.0e-9

    def test_attenuation_coefficient(self):
        """Test data is all at an air pressure of one standard atmosphere, 101.325 Pa."""
        data = np.loadtxt(
            data_path() / "absorption_coefficient.csv", skiprows=1, delimiter=","
        )
        f = np.array(
            [
                50.0,
                63.0,
                80.0,
                100.0,
                125.0,
                160.0,
                200.0,
                250.0,
                315.0,
                400.0,
                500.0,
                630.0,
                800.0,
                1000.0,
                1250.0,
                1600.0,
                2000.0,
                2500.0,
                3150.0,
                4000.0,
                5000.0,
                6300.0,
                8000.0,
                10000.0,
            ]
        )

        for row in data:
            temperature = 273.15 + row[0]  # Degrees Celsius to Kelvin
            relative_humidity = row[1]
            alpha = row[2:] / 1000.0  # Given in dB/km while we calculate in dB/m.

            assert f.shape == alpha.shape

            a = Atmosphere(temperature=temperature, relative_humidity=relative_humidity)
            calculated_alpha = a.attenuation_coefficient(f)

            np.testing.assert_array_almost_equal(alpha, calculated_alpha, decimal=2)



import numpy as np
from numpy.testing import assert_almost_equal

import pytest

from acoustic_toolbox.power import lw_iso3746


class TestPower:
    """Test :mod:`acoustic_toolbox.power`"""

    @pytest.mark.parametrize(
        "background_noise, expected",
        [
            (79, 91.153934187),
            (83, 90.187405234),
            (88, 88.153934187),
        ],
    )
    def test_lw_iso3746(self, background_noise, expected):
        LpAi = np.array([90, 90, 90, 90])
        LpAiB = background_noise * np.ones(4)
        S = 10
        alpha = np.array([0.1, 0.1, 0.1, 0.1, 0.1, 0.1])
        surfaces = np.array([10, 10, 10, 10, 10, 10])
        calculated = lw_iso3746(LpAi, LpAiB, S, alpha, surfaces)
        assert_almost_equal(calculated, expected)



"""The atmosphere module contains functions and classes related to atmospheric acoustics and is based on [ISO 9613-1:1993](standards/iso_9613_1_1993.md).

Classes:
    Atmosphere: Class describing atmospheric conditions.

Functions:
    soundspeed: Calculate the speed of sound.
    saturation_pressure: Calculate the saturation pressure.
    molar_concentration_water_vapour: Calculate the molar concentration of water vapour.
    relaxation_frequency_nitrogen: Calculate the relaxation frequency of nitrogen.
    relaxation_frequency_oxygen: Calculate the relaxation frequency of oxygen.
    attenuation_coefficient: Calculate the attenuation coefficient.
"""

from typing import Any
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from numpy.typing import NDArray

import acoustic_toolbox
from acoustic_toolbox.standards.iso_9613_1_1993 import (
    SOUNDSPEED,
    REFERENCE_TEMPERATURE,
    REFERENCE_PRESSURE,
    TRIPLE_TEMPERATURE,
    soundspeed,
    saturation_pressure,
    molar_concentration_water_vapour,
    relaxation_frequency_oxygen,
    relaxation_frequency_nitrogen,
    attenuation_coefficient,
)


class Atmosphere:
    """Class describing atmospheric conditions."""

    REF_TEMP = 293.15
    """Reference temperature"""

    REF_PRESSURE = 101.325
    """International Standard Atmosphere in kilopascal"""

    TRIPLE_TEMP = 273.16
    """Triple point isotherm temperature."""

    def __init__(
        self,
        temperature: float = REFERENCE_TEMPERATURE,
        pressure: float = REFERENCE_PRESSURE,
        relative_humidity: float = 0.0,
        reference_temperature: float = REFERENCE_TEMPERATURE,
        reference_pressure: float = REFERENCE_PRESSURE,
        triple_temperature: float = TRIPLE_TEMPERATURE,
    ):
        """Initialize atmosphere conditions.

        Args:
            temperature: Temperature in kelvin.
            pressure: Pressure.
            relative_humidity: Relative humidity.
            reference_temperature: Reference temperature.
            reference_pressure: Reference pressure.
            triple_temperature: Triple temperature.

        Returns:
            An instance of the Atmosphere class.
        """
        self.temperature = temperature
        """Ambient temperature $T$."""

        self.pressure = pressure
        """Ambient pressure $p_a$."""

        self.relative_humidity = relative_humidity
        """Relative humidity"""

        self.reference_temperature = reference_temperature
        """Reference temperature."""

        self.reference_pressure = reference_pressure
        """Reference pressure."""

        self.triple_temperature = triple_temperature
        """Triple temperature."""

    def __repr__(self) -> str:
        return "Atmosphere{}".format(self.__str__())

    def __str__(self) -> str:
        attributes = [
            "temperature",
            "pressure",
            "relative_humidity",
            "reference_temperature",
            "reference_pressure",
            "triple_temperature",
        ]
        return "({})".format(
            ", ".join(
                map(lambda attr: "{}={}".format(attr, getattr(self, attr)), attributes)
            )
        )

    def __eq__(self, other: Any) -> bool:
        return self.__dict__ == other.__dict__ and self.__class__ == other.__class__

    @property
    def soundspeed(self) -> float:
        """Speed of sound $c$.

        The speed of sound is calculated using [`standards.iso_9613_1_1993.soundspeed`][acoustic_toolbox.standards.iso_9613_1_1993.soundspeed].

        Returns:
            float: The speed of sound.
        """
        return soundspeed(
            self.temperature,
            self.reference_temperature,
        )

    @property
    def saturation_pressure(self) -> float:
        """Saturation pressure $p_{sat}$.

        The saturation pressure is calculated using [`standards.iso_9613_1_1993.saturation_pressure`][acoustic_toolbox.standards.iso_9613_1_1993.saturation_pressure].

        Returns:
            float: The saturation pressure.
        """
        return saturation_pressure(
            self.temperature,
            self.reference_pressure,
            self.triple_temperature,
        )

    @property
    def molar_concentration_water_vapour(self) -> float:
        """Molar concentration of water vapour $h$.

        The molar concentration of water vapour is calculated using
        [`standards.iso_9613_1_1993.molar_concentration_water_vapour`][acoustic_toolbox.standards.iso_9613_1_1993.molar_concentration_water_vapour].

        Returns:
            float: The molar concentration of water vapour.
        """
        return molar_concentration_water_vapour(
            self.relative_humidity,
            self.saturation_pressure,
            self.pressure,
        )

    @property
    def relaxation_frequency_nitrogen(self) -> float:
        """Resonance frequency of nitrogen $f_{r,N}$.

        The resonance frequency is calculated using
        [`standards.iso_9613_1_1993.relaxation_frequency_nitrogen`][acoustic_toolbox.standards.iso_9613_1_1993.relaxation_frequency_nitrogen].

        Returns:
            float: The resonance frequency of nitrogen.
        """
        return relaxation_frequency_nitrogen(
            self.pressure,
            self.temperature,
            self.molar_concentration_water_vapour,
            self.reference_pressure,
            self.reference_temperature,
        )

    @property
    def relaxation_frequency_oxygen(self) -> float:
        """Resonance frequency of oxygen $f_{r,O}$.

        The resonance frequency is calculated using
        [`standards.iso_9613_1_1993.relaxation_frequency_oxygen`][acoustic_toolbox.standards.iso_9613_1_1993.relaxation_frequency_oxygen].

        Returns:
            float: The resonance frequency of oxygen.
        """
        return relaxation_frequency_oxygen(
            self.pressure,
            self.molar_concentration_water_vapour,
            self.reference_pressure,
        )

    def attenuation_coefficient(
        self, frequency: float | NDArray[np.float64]
    ) -> float | NDArray[np.float64]:
        r"""Attenuation coefficient $\alpha$ describing atmospheric absorption in dB/m.

        The attenuation coefficient is calculated using
        [`standards.iso_9613_1_1993.attenuation_coefficient`][acoustic_toolbox.standards.iso_9613_1_1993.attenuation_coefficient].

        Args:
            frequency: Frequencies to be considered.

        Returns:
            float: The attenuation coefficient.
        """
        return attenuation_coefficient(
            self.pressure,
            self.temperature,
            self.reference_pressure,
            self.reference_temperature,
            self.relaxation_frequency_nitrogen,
            self.relaxation_frequency_oxygen,
            frequency,
        )

    def frequency_response(
        self,
        distance: float,
        frequencies: float | NDArray[np.float64],
        inverse: bool = False,
    ) -> NDArray[np.float64]:
        """Calculate the frequency response.

        Args:
            distance: Distance between source and receiver.
            frequencies: Frequencies for which to compute the response.
            inverse: Whether the attenuation should be undone.

        Returns:
            array: The frequency response.
        """
        return frequency_response(
            self,
            distance,
            frequencies,
            inverse,
        )

    def impulse_response(
        self,
        distance: float,
        fs: float,
        ntaps: int,
        inverse: bool = False,
    ) -> NDArray[np.float64]:
        """Calculate the impulse response of sound travelling through atmosphere.

        Args:
            distance: Distance between source and receiver.
            fs: Sample frequency.
            ntaps: Amount of taps.
            inverse: Whether the attenuation should be undone.

        Returns:
            array: The impulse response.

        See Also:
            [`atmosphere.impulse_response`][acoustic_toolbox.atmosphere.impulse_response]
        """
        return impulse_response(
            self,
            distance,
            fs,
            ntaps,
            inverse,
        )

    def plot_attenuation_coefficient(
        self, frequency: float | NDArray[np.float64]
    ) -> Figure:
        r"""Plot the attenuation coefficient $\alpha$ as function of frequency.

        Args:
            frequency: Frequencies.

        Note:
            The attenuation coefficient is plotted in dB/km!

        Returns:
            matplotlib.figure.Figure: The generated figure.
        """
        fig = plt.figure()
        ax0 = fig.add_subplot(111)
        ax0.plot(frequency, self.attenuation_coefficient(frequency) * 1000.0)
        ax0.set_xscale("log")
        ax0.set_yscale("log")
        ax0.set_xlabel(r"$f$ in Hz")
        ax0.set_ylabel(r"$\alpha$ in dB/km")
        ax0.legend()

        return fig


def frequency_response(
    atmosphere: Atmosphere,
    distance: float,
    frequencies: float | NDArray[np.float64],
    inverse: bool = False,
) -> NDArray[np.float64]:
    """Calculate the single-sided frequency response.

    Args:
        atmosphere: Atmosphere instance.
        distance: Distance between source and receiver.
        frequencies: Frequencies for which to compute the response.
        inverse: Whether the attenuation should be undone.

    Returns:
        array: The frequency response.
    """
    sign = +1 if inverse else -1
    tf = 10.0 ** (
        float(sign) * distance * atmosphere.attenuation_coefficient(frequencies) / 20.0
    )
    return tf


def impulse_response(
    atmosphere: Atmosphere,
    distance: float,
    fs: float,
    ntaps: int,
    inverse: bool = False,
) -> NDArray[np.float64]:
    """Calculate the impulse response of sound travelling through `atmosphere` for a given `distance` sampled at `fs`.

    The attenuation is calculated for a set of positive frequencies. Because the
    attenuation is the same for the negative frequencies, we have Hermitian
    symmetry. The attenuation is entirely real-valued. We like to have a constant
    group delay and therefore we need a linear-phase filter.

    This function creates a zero-phase filter, which is the special case of a
    linear-phase filter with zero phase slope. The type of filter is non-causal. The
    impulse response of the filter is made causal by rotating it by M/2 samples and
    discarding the imaginary parts. A real, even impulse response corresponds to a
    real, even frequency response.

    Args:
        atmosphere: Atmosphere instance.
        distance: Distance between source and receiver.
        fs: Sample frequency.
        ntaps: Amount of taps.
        inverse: Whether the attenuation should be undone.

    Returns:
        array: The impulse response.
    """
    # Frequencies vector with positive frequencies only.
    frequencies = np.fft.rfftfreq(ntaps, 1.0 / fs)
    # Single-sided spectrum. Negative frequencies have the same values.
    tf = frequency_response(atmosphere, distance, frequencies, inverse)
    # Impulse response. We design a zero-phase filter (linear-phase with zero slope).
    # We need to shift the IR to make it even. Taking the real part should not be necessary, see above.
    # ir = np.fft.ifftshift(np.fft.irfft(tf, n=ntaps)).real
    ir = acoustic_toolbox.signal.impulse_response_real_even(tf, ntaps=ntaps)
    return ir


__all__ = [
    "Atmosphere",
    "SOUNDSPEED",
    "REFERENCE_TEMPERATURE",
    "REFERENCE_TEMPERATURE",
    "TRIPLE_TEMPERATURE",
    "soundspeed",
    "saturation_pressure",
    "molar_concentration_water_vapour",
    "relaxation_frequency_oxygen",
    "relaxation_frequency_nitrogen",
    "attenuation_coefficient",
    "impulse_response",
    "frequency_response",
]



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal
import pytest

from acoustic_toolbox.building import (
    rw_curve,
    rw,
    rw_c,
    rw_ctr,
    stc,
    stc_curve,
    mass_law,
)


@pytest.fixture
def tl():
    return np.array([10, 11, 12, 13, 15, 18, 20, 20, 25, 22, 18, 15, 20, 25, 28, 30])


class TestBuilding:
    """Test :mod:`acoustic_toolbox.building`"""

    def test_rw_curve(self, tl):
        calculated = rw_curve(tl)
        real = np.array([3, 6, 9, 12, 15, 18, 21, 22, 23, 24, 25, 26, 26, 26, 26, 26])
        assert_array_almost_equal(calculated, real)

    def test_rw(self, tl):
        calculated = rw(tl)
        real = 22
        assert_almost_equal(calculated, real)

    def test_rw_c(self, tl):
        calculated = rw_c(tl)
        real = 19.657473180
        assert_almost_equal(calculated, real)

    def test_rw_ctr(self, tl):
        calculated = rw_ctr(tl)
        real = 18.036282792
        assert_almost_equal(calculated, real)

    def test_stc(self, tl):
        tl = np.array([18, 18, 17, 17, 20, 21, 22, 23, 25, 26, 26, 25, 25, 26, 30, 31])
        calculated = stc(tl)
        real = 25
        assert_almost_equal(calculated, real)

    def test_stc_curve(self, tl):
        tl = np.array([18, 18, 17, 17, 20, 21, 22, 23, 25, 26, 26, 25, 25, 26, 30, 31])
        calculated = stc_curve(tl)
        real = np.array([9, 12, 15, 18, 21, 24, 25, 26, 27, 28, 29, 29, 29, 29, 29, 29])
        assert_array_almost_equal(calculated, real)

    @pytest.mark.parametrize(
        "freq, vol_density, thickness, theta, c, rho0, expected",
        [
            (200, 2000, 0.005, 0, 343, 1.225, 23.514371494),
            (150, 8000, 0.002, 20, 345, 1.18, 24.827231087),
            (
                np.array([200, 150]),
                8000,
                0.002,
                20,
                345,
                1.18,
                np.array([27.319749023, 24.827231087]),
            ),
        ],
    )
    def test_mass_law(self, freq, vol_density, thickness, theta, c, rho0, expected):
        calculated = mass_law(freq, vol_density, thickness, theta, c, rho0)
        assert_array_almost_equal(calculated, expected)



"""Doppler shift module."""

SOUNDSPEED = 343.0
"""Speed of sound"""


def velocity_from_doppler_shift(f1, f2, c=SOUNDSPEED) -> float:
    r"""Calculate velocity based on measured frequency shifts due to Doppler shift.

    $$
    v = c \cdot \left( \frac{f_2 - f_1}{f_2 + f_1} \right)
    $$

    The assumption is made that the velocity is constant between the observation times.

    Args:
      f1: Lower frequency $f_1$.
      f2: Upper frequency $f_2$.
      c: Speed of sound $c$.

    Returns:
        Calculated velocity.
    """
    return c * (f2 - f1) / (f2 + f1)


def frequency_shift(
    frequency, velocity_source, velocity_receiver, soundspeed=SOUNDSPEED
) -> float:
    r"""Frequency shift due to Doppler effect.

    Args:
      frequency: Emitted frequency $f$.
      velocity_source: Velocity of source $v_s$.
        Positive if the source is moving away from the receiver (and negative in the other direction).
      velocity_receiver: Velocity of receiver $v_r$.
        Positive if the receiver is moving towards the source (and negative in the other direction);
      soundspeed: Speed of sound $c$.
        $$
        f = \frac{c + v_r}{c + v_s} f_0
        $$

    Returns:
        Frequency after shift.
    """
    return (soundspeed + velocity_receiver) / (soundspeed + velocity_source) * frequency



"""Sound power level calculations."""

import numpy as np


def lw_iso3746(
    LpAi: np.ndarray,
    LpAiB: np.ndarray,
    S: float,
    alpha: np.ndarray,
    surfaces: np.ndarray,
) -> float:
    r"""Calculate sound power level according to ISO 3746:2010.

    Args:
      LpAi: Sound pressure levels of the source $L_{pAi}$.
      LpAiB: Background noise sound pressure levels $L_{pAiB}$.
      S: Area in square meters of the measurement surface $S$.
      alpha: Absorption coefficients of the room $\alpha$.
      surfaces: Room surfaces.

    Returns:
      Sound power level $L_{w}$.
    """
    LpA = 10.0 * np.log10(np.sum(10.0 ** (0.1 * LpAi)) / LpAi.size)
    LpAB = 10.0 * np.log10(np.sum(10.0 ** (0.1 * LpAiB)) / LpAiB.size)
    deltaLpA = LpA - LpAB

    if deltaLpA > 10.0:
        k_1a = 0.0
    elif 3.0 <= deltaLpA <= 10.0:
        k_1a = -10.0 * np.log10(1.0 - 10.0 ** (-0.1 * deltaLpA))
    else:
        # This should alert to user because poor condition of the measurement.
        k_1a = 3.0

    S0 = 1.0
    Sv = np.sum(surfaces)
    alpha_mean = np.average(alpha, axis=0, weights=surfaces)
    A = alpha_mean * Sv

    k_2a = 10.0 * np.log10(1.0 + 4.0 * S / A)

    LpA_mean = LpA - k_1a - k_2a
    L_WA = LpA_mean + 10.0 * np.log10(S / S0)
    return L_WA



"""The descriptors module offers all kinds of acoustics related descriptors.

Descriptors from [ISO/TR 25417:2007](https://www.iso.org/standard/51150.html).

Attributes:
    REFERENCE_PRESSURE: Reference pressure for sound pressure level calculations.
    REFERENCE_SOUND_EXPOSURE: Reference sound exposure for sound exposure level calculations.
    REFERENCE_POWER: Reference power for sound power level calculations.
    REFERENCE_ENERGY: Reference energy for sound energy level calculations.
    REFERENCE_INTENSITY: Reference intensity for sound intensity level calculations.

Functions:
    sound_pressure_level: Calculate sound pressure level.
    equivalent_sound_pressure_level: Calculate equivalent sound pressure level.
    peak_sound_pressure: Calculate peak sound pressure.
    peak_sound_pressure_level: Calculate peak sound pressure level.
    sound_exposure: Calculate sound exposure.
    sound_exposure_level: Calculate sound exposure level.
    sound_power_level: Calculate sound power level.
    sound_energy: Calculate sound energy.
    sound_energy_level: Calculate sound energy level.
    sound_intensity: Calculate sound intensity.
    time_averaged_sound_intensity: Calculate time-averaged sound intensity.
    time_averaged_sound_intensity_level: Calculate time-averaged sound intensity level.
    normal_time_averaged_sound_intensity: Calculate normal time-averaged sound intensity.
    normal_time_averaged_sound_intensity_level: Calculate normal time-averaged sound intensity level.

Other descriptors
*****************

"""

import numpy as np

from acoustic_toolbox.standards.iso_tr_25417_2007 import (
    REFERENCE_PRESSURE,
    sound_pressure_level,
    equivalent_sound_pressure_level,
    peak_sound_pressure,
    peak_sound_pressure_level,
    REFERENCE_SOUND_EXPOSURE,
    sound_exposure,
    sound_exposure_level,
    REFERENCE_POWER,
    sound_power_level,
    sound_energy,
    REFERENCE_ENERGY,
    sound_energy_level,
    sound_intensity,
    time_averaged_sound_intensity,
    REFERENCE_INTENSITY,
    time_averaged_sound_intensity_level,
    normal_time_averaged_sound_intensity,
    normal_time_averaged_sound_intensity_level,
)

from acoustic_toolbox.standards.iso_1996_1_2003 import composite_rating_level


def _leq(levels, time):
    levels = np.asarray(levels)
    return 10.0 * np.log10((1.0 / time) * np.sum(10.0 ** (levels / 10.0)))


def leq(levels, int_time=1.0) -> float:
    """Equivalent level $L_{eq}$.

    Args:
      levels: Levels as function of time.
      int_time: Integration time in seconds.

    Returns:
      Equivalent level $L_{eq}$.
    """
    levels = np.asarray(levels)
    time = levels.size * int_time
    return _leq(levels, time)


def sel(levels: np.ndarray) -> float:
    """Sound Exposure Level from ``levels``."""
    levels = np.asarray(levels)
    return _leq(levels, 1.0)


def lw(W, Wref=1.0e-12) -> float:
    """Sound power level $L_{w}$ for sound power $W$ and reference power $W_{ref}$.

    Args:
      W: Sound power $W$.
      Wref: Reference power $W_{ref}$. Default value is $10^{12}$ watt.

    Returns:
      Sound power level $L_{w}$.
    """
    W = np.asarray(W)
    return 10.0 * np.log10(W / Wref)


def lden(
    lday,
    levening,
    lnight,
    hours: tuple[float, float, float] = (12.0, 4.0, 8.0),
    adjustment: tuple[float, float, float] = (0.0, 5.0, 10.0),
) -> float:
    """Calculate $L_{den}$ from $L_{day}$, $L_{evening}$ and $L_{night}$.

    Args:
      lday: Equivalent level during day period $L_{day}$.
      levening: Equivalent level during evening period $L_{evening}$.
      lnight: Equivalent level during night period $L_{night}$.
      hours: Hours per period.
      adjustment: Correction factor per period.

    Returns:
      $L_{den}$

    See Also:
      [composite_rating_level][acoustic_toolbox.standards.iso_1996_1_2003.composite_rating_level]

    """
    lday = np.asarray(lday)
    levening = np.asarray(levening)
    lnight = np.asarray(lnight)
    return composite_rating_level(
        np.vstack((lday, levening, lnight)).T, hours, adjustment
    )


def ldn(
    lday,
    lnight,
    hours: tuple[float, float] = (15.0, 9.0),
    adjustment: tuple[float, float] = (0.0, 10.0),
) -> float:
    """Calculate $L_{dn}$ from $L_{day}$ and $L_{night}$.

    Args:
      lday: Equivalent level during day period $L_{day}$.
      lnight: Equivalent level during night period $L_{night}$.
      hours: Hours per period.
      adjustment: Correction factor per period.

    Returns:
      $L_{dn}$

    See Also:
      [composite_rating_level][acoustic_toolbox.standards.iso_1996_1_2003.composite_rating_level]
    """
    lday = np.asarray(lday)
    lnight = np.asarray(lnight)
    return composite_rating_level(np.vstack((lday, lnight)).T, hours, adjustment)


__all__ = [
    # Following we take from another module
    "REFERENCE_PRESSURE",
    "sound_pressure_level",
    "equivalent_sound_pressure_level",
    "peak_sound_pressure",
    "peak_sound_pressure_level",
    "REFERENCE_SOUND_EXPOSURE",
    "sound_exposure",
    "sound_exposure_level",
    "REFERENCE_POWER",
    "sound_power_level",
    "sound_energy",
    "REFERENCE_ENERGY",
    "sound_energy_level",
    "sound_intensity",
    "time_averaged_sound_intensity",
    "REFERENCE_INTENSITY",
    "time_averaged_sound_intensity_level",
    "normal_time_averaged_sound_intensity",
    "normal_time_averaged_sound_intensity_level",
    # Following are locally defined
    "leq",
    "sel",
    "lw",
    "lden",
    "ldn",
]



"""The `decibel` module contains basic functions for decibel arithmetic."""

import numpy as np


def dbsum(levels, axis=None) -> float:
    r"""Energetic summation of levels.

    Args:
      levels: Sequence of levels.
      axis: Axis over which to perform the operation.
        $$
        L_{sum} = 10 \log_{10}{\sum_{i=0}^n{10^{L/10}}}
        $$

    Returns:
        Energetic summation of levels.
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0 ** (levels / 10.0)).sum(axis=axis))


def dbmean(levels, axis=None) -> float:
    r"""Energetic average of levels.

    Args:
      levels: Sequence of levels.
      axis: Axis over which to perform the operation.
        $$
        L_{mean} = 10 \log_{10}{\frac{1}{n}\sum_{i=0}^n{10^{L/10}}}
        $$

    Returns:
        Energetic average of levels.
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0 ** (levels / 10.0)).mean(axis=axis))


def dbadd(a: float | np.ndarray, b: float | np.ndarray) -> float:
    r"""Energetic addition.

    Energetically adds b to a.

      $$
      L_{a+b} = 10 \log_{10}{10^{L_b/10}+10^{L_a/10}}
      $$

    Args:
      a: Single level or sequence of levels.
      b: Single level or sequence of levels.


    Returns:
        Energetic addition of a and b.
    """
    a = np.asanyarray(a)
    b = np.asanyarray(b)
    return 10.0 * np.log10(10.0 ** (a / 10.0) + 10.0 ** (b / 10.0))


def dbsub(a: float | np.ndarray, b: float | np.ndarray) -> float:
    r"""Energetic subtraction.

    Energitally subtract b from a.

    $$
    L_{a-b} = 10 \log_{10}{10^{L_a/10}-10^{L_b/10}}
    $$

    Args:
      a: Single level or sequence of levels.
      b: Single level or sequence of levels.

    Returns:
        Energetic subtraction of a and b.
    """
    a = np.asanyarray(a)
    b = np.asanyarray(b)
    return 10.0 * np.log10(10.0 ** (a / 10.0) - 10.0 ** (b / 10.0))


def dbmul(levels, f, axis=None) -> np.ndarray:
    r"""Energetically add `levels` `f` times.

    $$
    L_{sum} = 10 \log_{10}{\sum_{i=0}^n{10^{L/10} \cdot f}}
    $$

    Args:
      levels: Sequence of levels.
      f: Multiplication factor `f`.
      axis: Axis over which to perform the operation.

    Returns:
        Resulting levels after energetic addition.
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0 ** (levels / 10.0) * f).sum(axis=axis))


def dbdiv(levels: float | np.ndarray, f: float, axis: int | None = None) -> np.ndarray:
    r"""Energetically divide `levels` `f` times.

    $$
    L_{sum} = 10 \log_{10}{\sum_{i=0}^n{10^{L/10} / f}}
    $$

    Args:
      levels: Sequence of levels.
      f: Divider `f`.
      axis: Axis over which to perform the operation.

    Returns:
        Resulting levels after energetic division.
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0 ** (levels / 10.0) / f).sum(axis=axis))


__all__ = ["dbsum", "dbmean", "dbadd", "dbsub", "dbmul", "dbdiv"]
